/*******************************************************************************
* Copyright (C) 2005, 2006 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
* Author: Jon Carey (jcarey@novell.com)
******************************************************************************/
#include <cmpiutil/base.h>
#include <cmpiutil/string.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <sys/utsname.h>

#include <cmpiutil/exec.h>
#include <stdlib.h>

//#define LOCAL_DEBUG

#ifdef LOCAL_DEBUG
#define DEBUGOUT(fmt, args...) printf(fmt,## args)
#else
#define DEBUGOUT(fmt, args...)
#endif



#define OSCLASSNAME "Linux_OperatingSystem"
#define CSCLASSNAME "Linux_ComputerSystem"
#define OSNAME "Linux"

/*****************************************************************************/
void
cmpiutilMilliSleep(
		uint32_t ms)
{
	struct timespec wait;
	wait.tv_sec = ms / 1000;
	wait.tv_nsec = (ms % 1000) * 1000000;
    while (nanosleep(&wait, &wait) == -1 && errno == EINTR)
	{
		/* Empty? */
	}
}

/*****************************************************************************/
void
cmpiutilSecondSleep(
		uint32_t secs)
{
	return cmpiutilMilliSleep(secs * 1000);
}


/*****************************************************************************/
const char*
cmpiutilGetOperatingSystemClassName(void)
{
	return OSCLASSNAME;
}

/*****************************************************************************/
const char*
cmpiutilGetComputerSystemClassName(void)
{
	return CSCLASSNAME;
}

/*****************************************************************************/
const char*
cmpiutilGetComputerSystemName(
		char* csNameOut,
		size_t csNameOutLen)
{
	if (!csNameOut || !csNameOutLen)
	{
		errno = EINVAL;
		return NULL;
	}
	return cmpiutilGetHostName(csNameOut, csNameOutLen);
}

/*****************************************************************************/
const char*
cmpiutilMakeInstanceID(
		const char* uniquePart,
		char* instanceIdOut,
		size_t instanceIdOutLen)
{
	if (!instanceIdOut || instanceIdOutLen < 5)
	{
		errno = EINVAL;
		return NULL;
	}

	*instanceIdOut = 0;
	cmpiutilStrAppend(instanceIdOut, "cmpiutil:", instanceIdOutLen);
	if (uniquePart)
	{
		cmpiutilStrAppend(instanceIdOut, uniquePart, instanceIdOutLen);
	}
	return instanceIdOut;
}

/*****************************************************************************/
const char*
cmpiutilGetHostName(
		char* hostNameOut,
		size_t hostNameOutLen)
{
	char namebuf[256];
	if (!hostNameOut || !hostNameOutLen)
	{
		errno = EINVAL;
		return NULL;
	}
	gethostname(namebuf, sizeof(namebuf));
	if (!strchr(namebuf, '.'))
	{
		char lbuf[2048];
		int herr = 0;
		struct hostent hostbuf;
		struct hostent* hent = &hostbuf;
		if (gethostbyname_r(namebuf, &hostbuf, lbuf, sizeof(lbuf),
			&hent, &herr) != 0)
		{
			hent = NULL;
		}

		if (hent && hent->h_name && strlen(hent->h_name) > 0)
		{
			cmpiutilStrNCpy(namebuf, hent->h_name, sizeof(namebuf));
		}
	}

	cmpiutilStrNCpy(hostNameOut, namebuf, hostNameOutLen);
	return hostNameOut;
}


/*****************************************************************************/
unsigned long long
cmpiutilGetRPMInstallDate(const char *rpmName)
{
	DEBUGOUT("Entering cmpiutilGetRPMInstallDate for rpmName: %s\n", rpmName);

	unsigned long rv = 0;
    int status;
    char *endptr, *output;
    size_t outputSize;

	const char *cmd[6];

	if (!rpmName || (*rpmName==0))
	{
		DEBUGOUT("  returning - empty rpmName\n");
		return 0;
	}

	cmd[0] = "/bin/rpm";
	cmd[1] = "-q";
	cmd[2] = "--queryformat";
	cmd[3] = "%{INSTALLTIME}";
	cmd[4] = rpmName;
	cmd[5] = NULL;  /* NULL termination */

	output = NULL;

	char *errOutput = NULL;
	size_t uErrOutputSize = 0;

	if (cmpiutilExecuteProcessAndGatherOutputAndError((char* const*) cmd, NULL, &output,
				&outputSize, -1, &errOutput, &uErrOutputSize, -1, NULL, &status, -1) == 0)
	{
		DEBUGOUT("  execute returned status: %d   outputSize: %d\n    output: %s\n", status, outputSize, output);
		if (status == 0)
		{
			rv = strtoul(output, &endptr, 10);
			if (endptr == output || *endptr)
			{
				rv = 0;
			}
			free(output);
			return (1000000 * (unsigned long long)rv);
		}
		else
		{
			DEBUGOUT("  received error:  %s\n", errOutput);
		}
	}
	if (output)
	{
		free(output);
	}

	return 0;
}

