/*******************************************************************************
* Copyright (C) 2005, 2006 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
* Author: Norm Paxton (npaxton@novell.com)
******************************************************************************/

#include "cmpiutil/cmpiTimeServiceUtils.h"
#include <cmpiutil/ntpServer.h>
#include <cmpiutil/ioUtils.h>
#include <fnmatch.h>
#include <ctype.h>

#define LOCAL_DEBUG

#ifdef LOCAL_DEBUG
#define DEBUGOUT(fmt, args...) printf(fmt,## args)
#else
#define DEBUGOUT(fmt, args...)
#endif


#ifdef __cplusplus
extern "C" {
#endif


// NULL terminated list of key property names for this class
static const char * systemTimeServiceClassKeys[] = {
			"SystemCreationClassName",
			"SystemName",
			"CreationClassName",
			"Name",
			NULL};


static const char * remoteTimeServicePortClassKeys[] = {
			"SystemCreationClassName",
			"SystemName",
			"CreationClassName",
			"Name",
			NULL};


static const char * timezoneSettingDataClassKeys[] = {
			"InstanceID",
			NULL};


/*****************************************************************************/
CMPIObjectPath *
makeTimeServiceObjectPath(
		const CMPIBroker *broker,
		const char *ns,
        CMPIStatus *pStatus)
{
	CMPIObjectPath *op = CMNewObjectPath(broker, ns, "Linux_SystemTimeService", pStatus);
	if (!CMIsNullObject(op) && (pStatus->rc == CMPI_RC_OK))
	{
		char name[128] = {0};
		cmpiutilGetComputerSystemName(name, sizeof(name));

		CMAddKey(op,
				"SystemCreationClassName",
				(CMPIValue *)(char *)cmpiutilGetComputerSystemClassName(),
				CMPI_chars);
		CMAddKey(op,
				"SystemName",
				(CMPIValue *)&name,
				CMPI_chars);
		CMAddKey(op,
				"CreationClassName",
				(CMPIValue *)"Linux_SystemTimeService",
				CMPI_chars);
		CMAddKey(op,
				"Name",
				(CMPIValue *)"cmpiutil:timeservice",
				CMPI_chars);
	}
	DEBUGOUT("makeTimeServiceObjectPath returning: %s\n", (char *)CMGetCharPtr(CMObjectPathToString(op, NULL)));
	return op;
}


/*****************************************************************************/
CMPIInstance *
makeTimeServiceInstance(
		const CMPIBroker *broker,
		const char *ns,
        CMPIStatus *pStatus,
        const char **properties)
{
	CMPIInstance *instance = NULL;
	CMPIObjectPath *op = makeTimeServiceObjectPath(broker, ns, pStatus);
	if (!CMIsNullObject(op) && (pStatus->rc == CMPI_RC_OK))
	{
		instance = CMNewInstance(broker, op, pStatus);
		if (!CMIsNullObject(instance) && (pStatus->rc == CMPI_RC_OK))
		{
			CMSetPropertyFilter(instance, properties, systemTimeServiceClassKeys);
			// keys
			char name[128] = {0};
			cmpiutilGetComputerSystemName(name, sizeof(name));

			CMSetProperty(instance,
					"SystemCreationClassName",
					(CMPIValue *)(char *)cmpiutilGetComputerSystemClassName(),
					CMPI_chars);
			CMSetProperty(instance,
					"SystemName",
					(CMPIValue *)&name,
					CMPI_chars);
			CMSetProperty(instance,
					"CreationClassName",
					(CMPIValue *)"Linux_SystemTimeService",
					CMPI_chars);
			CMSetProperty(instance,
					"Name",
					(CMPIValue *)"cmpiutil:timeservice",
					CMPI_chars);

			unsigned int enabledState = 2;
			CMSetProperty(instance, "EnabledState", (CMPIValue*)&enabledState, CMPI_uint16);
			unsigned int requestedState = 12;
			CMSetProperty(instance, "RequestedState", (CMPIValue*)&requestedState, CMPI_uint16);
			unsigned int enabledDefault = 2;
			CMSetProperty(instance, "EnabledDefault", (CMPIValue*)&enabledDefault, CMPI_uint16);
			unsigned int started = 1;
			CMSetProperty(instance, "Started", (CMPIValue*)&started, CMPI_boolean);
			unsigned int healthState = 5;
			CMSetProperty(instance, "HealthState", (CMPIValue*)&healthState, CMPI_uint16);

			CMPIDateTime *lastStateChange = CMNewDateTime(broker, NULL);
			CMSetProperty(instance, "TimeOfLastStateChange", (CMPIValue*)&lastStateChange, CMPI_dateTime);


			// TODO: Get the real one, move LinuxOS_get*** into common library?
			//       get from the linux os instance?
			//CMPIDateTime *installDate = CMNewDateTimeFromBinary(broker, cmpiutilLinuxOS_getInstallDate(), 0, NULL);
			CMPIDateTime *installDate = CMNewDateTime(broker, NULL);
			CMSetProperty(instance, "InstallDate", (CMPIValue*)&installDate, CMPI_dateTime);

			CMSetProperty(instance, "Status", (CMPIValue*)"OK", CMPI_chars);
			CMSetProperty(instance, "StartMode", (CMPIValue*)"Automatic", CMPI_chars);

			CMPIArray *pra = CMNewArray(broker, 1, CMPI_uint16, NULL);
			if (!CMIsNullObject(pra))
			{
				unsigned int opStatus = 2;
				CMSetArrayElementAt(pra, 0, (CMPIValue*)&opStatus, CMPI_uint16);
				CMSetProperty(instance, "OperationalStatus", (CMPIValue*)&pra, CMPI_uint16A);
			}
		}
	}
	return instance;
}



/*****************************************************************************/
CMPIObjectPath*
makeRemoteTimeServicePortObjectPath(
	const CMPIBroker *broker,
	const char *ns,
	const char *serverName,
	CMPIStatus *pStatus)
{
	CMPIObjectPath *op = CMNewObjectPath(broker, ns, "Linux_RemoteTimeServicePort", pStatus);
	if (!CMIsNullObject(op) && (pStatus->rc == CMPI_RC_OK))
	{
		char name[128] = {0};
		cmpiutilGetComputerSystemName(name, sizeof(name));

		CMAddKey(op,
				"SystemCreationClassName",
				(CMPIValue *)(char *)cmpiutilGetComputerSystemClassName(),
				CMPI_chars);
		CMAddKey(op,
				"SystemName",
				(CMPIValue *)&name,
				CMPI_chars);
		CMAddKey(op,
				"CreationClassName",
				(CMPIValue *)"Linux_RemoteTimeServicePort",
				CMPI_chars);
		CMAddKey(op,
				"Name",
				(CMPIValue *)serverName,
				CMPI_chars);
	}
	return op;
}


/*****************************************************************************/
CMPIInstance *
makeRemoteTimeServicePortInstance(
		const CMPIBroker *broker,
		const char *ns,
		const char *serverName,
        CMPIStatus *pStatus,
        const char **properties)
{
	DEBUGOUT("  entering makeRemoteTimeServicePortInstance\n");

	CMPIInstance *instance = NULL;
	CMPIObjectPath *op = makeRemoteTimeServicePortObjectPath(broker, ns, serverName, pStatus);
	if (!CMIsNullObject(op) && (pStatus->rc == CMPI_RC_OK))
	{
		DEBUGOUT("    about to getNTPServer\n");
		NTPServer *server = getNTPServer(serverName);
		if (server)
		{
			DEBUGOUT("    got a server\n");
			instance = CMNewInstance(broker, op, pStatus);
			if (!CMIsNullObject(instance) && (pStatus->rc == CMPI_RC_OK))
			{
				DEBUGOUT("    created instance\n");
				CMSetPropertyFilter(instance, properties, remoteTimeServicePortClassKeys);
				// keys
				char name[128] = {0};
				cmpiutilGetComputerSystemName(name, sizeof(name));

				CMSetProperty(instance,
						"SystemCreationClassName",
						(CMPIValue *)(char *)cmpiutilGetComputerSystemClassName(),
						CMPI_chars);
				CMSetProperty(instance,
						"SystemName",
						(CMPIValue *)&name,
						CMPI_chars);
				CMSetProperty(instance,
						"CreationClassName",
						(CMPIValue *)"Linux_RemoteTimeServicePort",
						CMPI_chars);
				CMSetProperty(instance,
						"Name",
						(CMPIValue *)serverName,
						CMPI_chars);

				CMSetProperty(instance,
						"AccessInfo",
						(CMPIValue *)serverName,
						CMPI_chars);

				DEBUGOUT("    done setting keys\n");

				int infoFmt = (isIPV4Address(serverName)) ? 3 : 2;
				CMSetProperty(instance, "InfoFormat", (CMPIValue*)&infoFmt, CMPI_uint16);

				unsigned int enabledState = 2;
				CMSetProperty(instance, "EnabledState", (CMPIValue*)&enabledState, CMPI_uint16);

				unsigned int requestedState = 12;
				CMSetProperty(instance, "RequestedState", (CMPIValue*)&requestedState, CMPI_uint16);

				unsigned int enabledDefault = 2;
				CMSetProperty(instance, "EnabledDefault", (CMPIValue*)&enabledDefault, CMPI_uint16);

				unsigned int started = 1;
				CMSetProperty(instance, "Started", (CMPIValue*)&started, CMPI_boolean);

				unsigned int healthState = 5;
				CMSetProperty(instance, "HealthState", (CMPIValue*)&healthState, CMPI_uint16);

				CMPIDateTime *lastStateChange = CMNewDateTime(broker, NULL);
				CMSetProperty(instance, "TimeOfLastStateChange", (CMPIValue*)&lastStateChange, CMPI_dateTime);

				DEBUGOUT("    done setting first set of vals\n");

				// TODO: Get the real one, move LinuxOS_get*** into common library?
				//       get from the linux os instance?
				//CMPIDateTime *installDate = CMNewDateTimeFromBinary(broker, cmpiutilLinuxOS_getInstallDate(), 0, NULL);
				CMPIDateTime *installDate = CMNewDateTime(broker, NULL);
				CMSetProperty(instance, "InstallDate", (CMPIValue*)&installDate, CMPI_dateTime);

				CMSetProperty(instance, "Status", (CMPIValue*)"OK", CMPI_chars);
				CMSetProperty(instance, "StartMode", (CMPIValue*)"Automatic", CMPI_chars);

				CMPIArray *pra = CMNewArray(broker, 1, CMPI_uint16, NULL);
				if (!CMIsNullObject(pra))
				{
					unsigned int opStatus = 2;
					CMSetArrayElementAt(pra, 0, (CMPIValue*)&opStatus, CMPI_uint16);
					CMSetProperty(instance, "OperationalStatus", (CMPIValue*)&pra, CMPI_uint16A);
				}

				unsigned int portProtocol = 2;
				CMSetProperty(instance, "PortProtocol", (CMPIValue*)&portProtocol, CMPI_uint16);

				CMPIBoolean prefer = 0;
				prefer = (CMPIBoolean)(server->prefer);
				CMSetProperty(instance, "Prefer", (CMPIValue*)&prefer, CMPI_boolean);

				DEBUGOUT("    setting instance key: %d\n", server->key);
				DEBUGOUT("    setting instance prefer: %d\n", server->prefer);
				DEBUGOUT("    setting instance ntpversion: %d\n", server->ntpVersion);
				DEBUGOUT("    setting instance minpoll: %d\n", server->minPoll);
				DEBUGOUT("    setting instance maxpoll: %d\n", server->maxPoll);
				if(server->key)
					CMSetProperty(instance, "Key", (CMPIValue*)&(server->key), CMPI_uint32);
				if(server->ntpVersion)
					CMSetProperty(instance, "NTPversion", (CMPIValue*)&(server->ntpVersion), CMPI_uint32);
				if(server->minPoll)
					CMSetProperty(instance, "MinPoll", (CMPIValue*)&(server->minPoll), CMPI_uint32);
				if(server->maxPoll)
					CMSetProperty(instance, "MaxPoll", (CMPIValue*)&(server->maxPoll), CMPI_uint32);

				DEBUGOUT("    done setting instance properties\n");
			}
			free(server);
		}
	}
	DEBUGOUT("  leaving makeRemoteTimeServicePortInstance\n");
	return instance;
}



/*****************************************************************************/
CMPIObjectPath*
makeTimeZoneSettingDataObjectPath(
		const CMPIBroker *broker,
		const char *ns,
		CMPIStatus *pStatus)
{
	CMPIObjectPath *op = CMNewObjectPath(broker, ns, "Linux_TimeZoneSettingData", pStatus);
	if (!CMIsNullObject(op) && (pStatus->rc == CMPI_RC_OK))
	{
		char name[128] = {0};
		cmpiutilGetComputerSystemName(name, sizeof(name));

		CMAddKey(op,
				"InstanceID",
				(CMPIValue *)"cmpiutil:timezone",
				CMPI_chars);
	}
	return op;
}



/*****************************************************************************/
CMPIInstance *
makeTimeZoneSettingDataInstance(
		const CMPIBroker *broker,
		const char *ns,
        CMPIStatus *pStatus,
        const char **properties)
{
	CMPIInstance *instance = NULL;
	CMPIObjectPath *op = makeTimeZoneSettingDataObjectPath(broker, ns, pStatus);
	if (!CMIsNullObject(op) && (pStatus->rc == CMPI_RC_OK))
	{
		char timezone[32] = {0};
		int utcHwClock = 1;

		cmpiutilGetClockInfo(timezone, 32, &utcHwClock);

		instance = CMNewInstance(broker, op, pStatus);
		if (!CMIsNullObject(instance) && (pStatus->rc == CMPI_RC_OK))
		{
			CMSetPropertyFilter(instance, properties, timezoneSettingDataClassKeys);
			// keys
			CMSetProperty(instance,
					"InstanceID",
					(CMPIValue *)"cmpiutil:timezone",
					CMPI_chars);

			CMSetProperty(instance,
					"TimeZone",
					(CMPIValue *)timezone,
					CMPI_chars);

			CMSetProperty(instance, "UTCHardwareClock", (CMPIValue*)&utcHwClock, CMPI_boolean);
		}
	}
	return instance;
}



/*****************************************************************************/
int
isIPV4Address(const char *arg)
{
	const char *p = arg;
	if(fnmatch("*.*.*.*", p, 0) != 0)
	{
		return 0;
	}

	for(; *p; p++)
	{
		if(*p != '.' && !isdigit(*p))
		{
			return 0;
		}
	}

	return 1;
}


/*****************************************************************************/
static int
getKeyValuePair(
		const char *line,
		char *key,
		int keylen,
		char *value,
		int valuelen)
{
	int retval = 0;
	char *delim = strchr(line, '=');
	if(delim)
	{
		int localKeyLen = (int)((char*)delim-(char*)line+1);
		int localValLen = (int)((char*)(line+strlen(line))-(char*)delim);
		cmpiutilStrNCpy(key, line, (localKeyLen > keylen?keylen:localKeyLen));
		cmpiutilStrTrim(key);
		cmpiutilStrNCpy(value, delim+1, (localValLen > valuelen?valuelen:localValLen));
		cmpiutilStrTrim(value);
		if (value[0] == '"')
		{
			cmpiutilStrNCpy(value, value+1, strlen(value)-1 );
		}
		if (cmpiutilStrEndsWith(value, "\""))
		{
			value[strlen(value)-1] = '\0';
		}
		retval = 1;
	}

	return 0;
}

/*****************************************************************************/
void cmpiutilGetClockInfo(
		char *timezone,
		int timezoneLen,
		int *utcHwClock)
{
	FILE *file;
	size_t bfrLen = 0;
	char *p = NULL, *bufPtr = NULL;
	//char* errMsgBfr = NULL;
	//unsigned int errMsgBfrLen = 0;

	DEBUGOUT("Entering cmpiutilGetClockInfo\n");

	if (!timezone || !utcHwClock || !timezoneLen)
	{
		return;
	}

	file = fopen("/etc/sysconfig/clock", "r");
	if (!file)
	{
		//errCond(errMsgBfr, errMsgBfrLen, -1,
		//	"Failed to open file /etc/sysconfig/clock");
		return;
	}

	char key[32] = {0};
	char value[256] = {0};

	for(p = cmpiutilReadLine(file, &bufPtr, &bfrLen); p;
		p = cmpiutilReadLine(file, &bufPtr, &bfrLen))
	{
		DEBUGOUT("  got line: %s\n", bufPtr);

		if (bufPtr[0] != '#')
		{
			memset(key, 0, 32);
			memset(value, 0, 256);
			getKeyValuePair(bufPtr, key, 32, value, 256);
			DEBUGOUT("  Got Key/Value pair: %s : %s\n", key, value);
			if (strcasecmp(key, "HWCLOCK") == 0)
			{
				DEBUGOUT("    Found HWCLOCK\n");
				if (strcasecmp(value, "-u") == 0)
				{
					DEBUGOUT("      setting to 1\n");
					*utcHwClock = 1;
				}
				else
				{
					DEBUGOUT("      setting to 0\n");
					*utcHwClock = 0;
				}
			}
			else if (strcasecmp(key, "TIMEZONE") == 0)
			{
				DEBUGOUT("    Found TIMEZONE: %s\n", value);
				cmpiutilStrNCpy(timezone, value, timezoneLen);
				DEBUGOUT("    returning timezone: %s\n", timezone);
			}
		}
	}

	fclose(file);

	if (bufPtr)
		free(bufPtr);
}


/*****************************************************************************/
int
cmpiutilSetClockInfo(
		char *timezone,
		int utcHwClock)
{
	// TODO:  Write this
	return 1;
}
