/*******************************************************************************
* Copyright (C) 2005, 2006 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************/


#include "cmpiutil/fileTools.h"
#include <cmpiutil/string.h>
#include <cmpiutil/ioUtils.h>

#include <libgen.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdlib.h>
#include <stdio.h>

char* cmpiutil_dirname(const char* path)
{
	char* tmp = strdup(path);
	char* rv = dirname(tmp);
	if (rv == NULL)
	{
		free(tmp);
	}
	return rv;
}

void cmpiutilFileKeys_free(cmpiutilFileKeys* fk)
{
	if (fk == NULL)
	{
		return;
	}
	//free(fk->CSCreationClassName); // not malloc'd
	free(fk->CSName);
	//free(fk->CreationClassName); // not malloc'd
	free(fk->FSCreationClassName);
	free(fk->FSName);
	free(fk->Name);
	free(fk);
}

cmpiutilFileKeys* cmpiutilGetKeysForFile(const char* path)
{
	struct stat st;
	if (lstat(path, &st) != 0)
	{
		return NULL;
	}
	return cmpiutilGetKeysForFileStat(path, &st);
}

cmpiutilFileKeys* cmpiutilGetKeysForFileStat(const char* path, struct stat* st)
{
	cmpiutilFileKeys* rv = (cmpiutilFileKeys*)malloc(sizeof(cmpiutilFileKeys));
	memset(rv, 0, sizeof(cmpiutilFileKeys));
	if (S_ISREG(st->st_mode))
	{
		rv->CreationClassName = "Linux_LinuxDataFile";
	} else if (S_ISDIR(st->st_mode))
	{
		rv->CreationClassName = "Linux_LinuxDirectory";
	} else if (S_ISCHR(st->st_mode) || S_ISBLK(st->st_mode))
	{
		rv->CreationClassName = "Linux_LinuxDeviceFile";
	} else if (S_ISFIFO(st->st_mode))
	{
		rv->CreationClassName = "Linux_LinuxFIFOPipeFile";
	} else if (S_ISLNK(st->st_mode))
	{
		rv->CreationClassName = "Linux_LinuxSymbolicLink";
	} else
	{
		free(rv);
		return NULL;
	}
	rv->Name = strdup(path);
	rv->CSCreationClassName = "Linux_ComputerSystem";
	rv->CSName = (char*)malloc(256);
	gethostname(rv->CSName, 256);
	FILE* mounts = fopen("/proc/mounts", "r");
	if (mounts == NULL)
	{
		cmpiutilFileKeys_free(rv);
		return NULL;
	}
	char* line = NULL;
	char* p = NULL;
	size_t bfrLen = 0;
	for(p = cmpiutilReadLine(mounts, &line,&bfrLen); p; p = cmpiutilReadLine(mounts, &line,&bfrLen ))
	{
		int count = cmpiutilStrTokenCount(line, " \t\r\n");
		if (count != 6)
		{
			continue;
		}
		size_t numElems = 0;
		char** strs = cmpiutilStrTokenize(line, " \t\r\n", &numElems);
		if (strs == NULL)
		{
			continue;
		}
		if (strs[1][0] != '/')
		{
			free(strs);
			continue;
		}
		if (strs[1][1] == 0 && strcmp(strs[0], "rootfs") == 0)
		{
			/* skip rootfs junk */
			free(strs);
			continue;
		}
		struct stat dst;
		if (stat(strs[1], &dst) != 0)
		{
			free(strs);
			continue;
		}
		if (st->st_dev == dst.st_dev)
		{
			rv->FSName = strdup(strs[0]);
			char* fstype = strs[2];
			if (strcmp(fstype, "ext2") == 0)
			{
				rv->FSCreationClassName = strdup("Linux_Ext2FileSystem");
			} else if (strcmp(fstype, "ext3") == 0)
			{
				rv->FSCreationClassName = strdup("Linux_Ext3FileSystem");
			} else if (strcmp(fstype, "reiserfs") == 0)
			{
				rv->FSCreationClassName = strdup("Linux_ReiserFileSystem");
			} else if (strcmp(fstype, "nfs") == 0)
			{
				rv->FSCreationClassName = strdup("Linux_NFS");
			} else if (strcmp(fstype, "cifs") == 0)
			{
				rv->FSCreationClassName = strdup("Linux_CIFS");	/* TODO CIM class doesn't exist yet */
			}
			else
			{
				int len = snprintf(NULL, 0, "Unknown:%s", fstype);
				if (len > 0)
				{
					rv->FSCreationClassName = (char*)malloc(len + 1);
					snprintf(rv->FSCreationClassName, len+1, "Unknown:%s", fstype);
				}
				else
				{
					rv->FSCreationClassName = strdup("Unknown");
				}
			}
		}
		free(strs);
		if (rv->FSName != NULL)
		{
			break;
		}
	}
	if (line)
	{
		free(line);
	}
	fclose(mounts);
	return rv;

}

