/*******************************************************************************
* Copyright (C) 2005, 2006 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
* Author: Jon Carey (jcarey@novell.com)
******************************************************************************/
#include <cmpiutil/ioUtils.h>
#include <cmpiutil/string.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>

/*****************************************************************************/
FILE*
cmpiutilCreateTempFileInDir(
	const char* dirName,
	char* fileNameOut,
	size_t fileNameOutLen)
{
	char lfbfr[PATH_MAX];
	int lerrno, fd;
	uint8_t autoClose = 0;
	char *penv;
	FILE *fhdl;

	if (!fileNameOut)
	{
		fileNameOut = lfbfr;
		fileNameOutLen = sizeof(lfbfr);
		autoClose = 1;
	}

	if (dirName && *dirName) 
	{
		cmpiutilStrNCpy(fileNameOut, dirName, fileNameOutLen);
		if (!cmpiutilStrEndsWith(fileNameOut, "/"))
			cmpiutilStrAppend(fileNameOut, "/", fileNameOutLen);
	}
	else 
	{
		penv = getenv("TMPDIR");
		if (penv) 
		{
			cmpiutilStrNCpy(fileNameOut, penv, fileNameOutLen);
			if (!cmpiutilStrEndsWith(fileNameOut, "/"))
				cmpiutilStrAppend(fileNameOut, "/", fileNameOutLen);
		}
		else 
		{
			cmpiutilStrNCpy(fileNameOut, "/tmp/", fileNameOutLen);
		}
	}
	cmpiutilStrAppend(fileNameOut, "cmpiutiltmpfileXXXXXX", fileNameOutLen);

	fd = mkstemp(fileNameOut);
	if (fd == -1) 
	{
		return NULL;
	}

	fhdl = fdopen(fd, "w+");
	if (!fhdl) 
	{
		printf("Inside 'if'\n");
		lerrno = errno;
		close(fd);
		unlink(fileNameOut);
		errno = lerrno;
		return NULL;
	}

	if (autoClose) 
	{
		/* The caller doesn't want the file name returned, so we'll
		 * unlink it now. This will cause it to get deleted as soon
		 * as the file is closed.
		 */
		unlink(fileNameOut);
	}
	return fhdl;
}

/*****************************************************************************/
FILE*
cmpiutilCreateTempFile(
	char* fileNameOut,
	size_t fileNameOutLen)
{
	return cmpiutilCreateTempFileInDir(0, fileNameOut, fileNameOutLen);
}

/*****************************************************************************/
#define INIT_READ_LINE_LEN 128
#define INC_READ_LINE_LEN 64

char*
cmpiutilReadLine(FILE* fd, char **bfr, size_t* const bfrLen)
{
	char *p, *lnptr, *lbfr, *rcptr = NULL;
	int readLen;

	if (!(*bfr)) 
	{
		*bfr = (char*) malloc(INIT_READ_LINE_LEN);
		if (!(*bfr)) 
		{
			return NULL;
		}
		*bfrLen = INIT_READ_LINE_LEN;
	}
	*bfr[0] = 0;

	readLen = *bfrLen;
	lnptr = *bfr;
	for(;;) 
	{
		p = fgets(lnptr, readLen, fd);
		if (!p) 
		{
			/* Assume EOF condition. If lnptr still points to the beginning
			 * of the buffer, then return a null. Otherwise we'll return a
			 * pointer to the buffer, because we have retrieved data.
			 */
			rcptr = (lnptr == *bfr) ? NULL : *bfr;
			break;
		}
		if ((p = strchr(*bfr, '\n'))) 
		{
			/* Read an EOL so exit function */
			*p = '\0';
			rcptr = *bfr;
			break;
		}

		/* At this point we know we haven't read an entire line,
		 * so we'll assume that the buffer wasn't long enough.
		 * We'll realloc for more memory, and continue to
		 * read until we do get the EOL.
		 */
		lbfr = (char*) realloc(*bfr, *bfrLen + INC_READ_LINE_LEN);
		if (!lbfr) 
		{
			return NULL;
		}
		*bfr = lbfr;
		lnptr = *bfr + *bfrLen - 1;
		*bfrLen += INC_READ_LINE_LEN;
		readLen = INC_READ_LINE_LEN+1;
	}

	return rcptr;
}

