#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <regex.h>

#include <cmpiutil/ioUtils.h>
#include <cmpiutil/string.h>
#include <cmpiutil/linuxProcUtils.h>

static time_t _system_boot_time = 0;
static int _getTaskState(pid_t pid, int tid);
static enum ProcState _stateThruTasks(pid_t pid);
static FILE* _openStatFile(pid_t pid);
static FILE* _openMemFile(pid_t pid);
static int _appendPid2Array(pid_t** intBfr, size_t* intBfrSize,
	size_t* offset, pid_t value);

/*****************************************************************************/
time_t
cmpiutilGetSystemBootTime(void)
{
	if(_system_boot_time == 0) 
	{
		FILE* sfdl = fopen("/proc/stat", "r");
		if(sfdl) 
		{
			char *p, *bfr = NULL;
			size_t bfrLen = 0;
			for(p = cmpiutilReadLine(sfdl, &bfr, &bfrLen); p;
				p = cmpiutilReadLine(sfdl, &bfr, &bfrLen)) 
			{
				if(!strncmp(bfr, "btime ", 6)) 
				{
					_system_boot_time = strtoul(p+6, NULL, 10);
					break;
				}
			}
			fclose(sfdl);
			if (bfr) 
			{
				free(bfr);
			}
		}
	}

	return _system_boot_time;
}

/*****************************************************************************/
enum ProcState
cmpiutilGetProcState(pid_t pid)
{
	return _stateThruTasks(pid);
}

/*****************************************************************************/
int
cmpiutilValidPidFile(
	const char* pidFileName,
	const char* execName)
{
	int cc = 0;
	pid_t pid;
	if(access(pidFileName, R_OK) == 0) 
	{
		FILE* fd = fopen(pidFileName, "r");
		if(fd) 
		{
			char lbfr[1024];
			const char* p;
			fscanf(fd, "%d", &pid);
			fclose(fd);
			if ((p = cmpiutilGetProcExecName(pid, lbfr, sizeof(lbfr)))) 
			{
				if (!strcmp(execName, lbfr)) 
				{
					if (cmpiutilGetProcState(pid) != EPROCSTATE_DoesntExist) 
					{
						cc = 1;
					}
				}
			}
		}
	}
	return cc;
}

/*****************************************************************************/
const char*
cmpiutilGetProcExecName(
	pid_t pid,
	char* execNameOut,
	size_t execNameOutLen)
{
	FILE* fd;
	char bfr[1024];
	const char* fmt =
		"%*d "		/* pid %d */
		"%1023s "	/* comm %s - Size of bfr - 1 */
		"%*c "		/* state %c */
		"%*d "		/* ppid %d */
		"%*d "		/* pgrp %d */
		"%*d "		/* session %d */
		"%*d "		/* tty nr %d */
		"%*d "		/* tty pgrp %d */
		"%*lu "		/* flags %lu */
		"%*lu "		/* min flt %lu */
		"%*lu "		/* cmin flt %lu */
		"%*lu "		/* maj flt %lu */
		"%*lu "		/* cmaj flt %lu */
		"%*lu "		/* utime %lu */
		"%*lu "		/* stime %lu */
		"%*ld "		/* cutime %ld */
		"%*ld "		/* cstime %ld */
		"%*ld "		/* priority %ld */
		"%*ld "		/* nice %ld */
		"%*ld "		/* removed %ld */
		"%*ld "		/* itrealvalue %ld */
		"%*lu "		/* start time %lu */
		"%*lu "		/* vsize %lu */
		"%*ld "		/* rss %ld */
		"%*lu "		/* rlim %lu */
		"%*lu "		/* start code %lu */
		"%*lu "		/* end code %lu */
		"%*lu "		/* start stack %lu */
		"%*lu "		/* esp %lu */
		"%*lu "		/* eip %lu */
		"%*lu "		/* signal %lu */
		"%*lu "		/* blocked %lu */
		"%*lu "		/* sigignore %lu */
		"%*lu "		/* sigcatch %lu */
		"%*lu "		/* wchan %lu */
		"%*lu "		/* nswap %lu */
		"%*lu "		/* cnswap %lu */
		"%*d "		/* exit signal %d */
		"%*d";		/* cpu number %d */

	if (!execNameOut || !execNameOutLen) 
	{
		errno = EINVAL;
		return NULL;
	}

	fd = _openStatFile(pid);
	if(fd) 
	{
		fscanf(fd, fmt, bfr);
		fclose(fd);
		bfr[strlen(bfr)-1] = 0;
		cmpiutilStrNCpy(execNameOut, &bfr[1], execNameOutLen);
		return execNameOut;
	}

	return NULL;
}

/*****************************************************************************/
int
cmpiutilGetAllProcs(
	pid_t **procsArg,
	size_t *numProcs,
	const char* regexPattern)
{
	DIR* pdir;
	struct dirent* pdirent;
	regex_t re;
	int cc, isRegEx = 0;
	int *lprocs = 0;
	size_t intBfrSize = 0;
	size_t offset = 0;

	if (procsArg) 
	{
		*procsArg = 0;
	}

	if (numProcs) 
	{
		*numProcs = 0;
	}

	pdir = opendir("/proc");
	if(!pdir) 
	{
		return -1;
	}

	if(regexPattern && *regexPattern) 
	{
		isRegEx = 1;
		if(regcomp(&re, regexPattern, REG_EXTENDED|REG_NOSUB) != 0) 
		{
			closedir(pdir);
			return -1;
		}
	}

	cc = 0;
	while(((pdirent = readdir(pdir)) != NULL) && cc == 0) 
	{
		char *endptr;
		long lpid = strtol(pdirent->d_name, &endptr, 10);
		if(endptr == pdirent->d_name || *endptr) 
		{
			continue;
		}
		if(isRegEx) 
		{
			/* Regex pattern specified, so make sure the exec name of
			 * the process matches the given regular expression. */
			char lbfr[1024];
			const char* p = cmpiutilGetProcExecName((pid_t)lpid, lbfr, sizeof(lbfr));
			if (p) 
			{
				if(regexec(&re, p, (size_t)0, NULL, 0) == 0) 
				{
					if (procsArg) 
					{
						if (_appendPid2Array(&lprocs, &intBfrSize, &offset,
							(pid_t)lpid) == -1) 
						{
							cc = -1;
						}
					}
					else 
					{
						offset++;
					}
				}
			}
		}
		else 
		{
			if (procsArg) 
			{
				if (_appendPid2Array(&lprocs, &intBfrSize, &offset,
					(pid_t)lpid) == -1) 
				{
					cc = -1;
				}
			}
			else 
			{
				offset++;
			}
		}
	}
	closedir(pdir);
	if(isRegEx) 
	{
		regfree(&re);
	}

	if (procsArg) 
	{
		*procsArg = lprocs;	/* Caller must free this buffer */
	}

	if (numProcs) 
	{
		*numProcs = offset;	/* Return number of procs found */
	}
	return 0;
}

/*****************************************************************************/
uint64_t
cmpiutilGetProcRealUID(
	pid_t pid)
{
	uint64_t ruid = (uint64_t)-1;
	char fname[32];
	FILE* fhdl;

	snprintf(fname, sizeof(fname), "/proc/%d/status", pid);
	fhdl = fopen(fname, "r");
	if(fhdl)
	{
		char *p, *bfr = NULL;
		size_t bfrLen = 0;
		for(p = cmpiutilReadLine(fhdl, &bfr, &bfrLen); p;
			p = cmpiutilReadLine(fhdl, &bfr, &bfrLen)) 
		{
			if(!strncmp(bfr, "Uid:", 4)) 
			{
				sscanf(bfr, "Uid: %llu", &ruid);
				break;
			}
		}
		fclose(fhdl);
		if (bfr) 
		{
			free(bfr);
		}
	}
	return ruid;
}

/*****************************************************************************/
#define HERTZ 100
double
cmpiutilGetProcPercentCPU(
	pid_t pid,
	int includeDeadChildren)
{
	const char* fmt =
		"%*d "		// pid %d
		"%*s "		// comm %s
		"%*c "		// state %c
		"%*d "		// ppid %d
		"%*d "		// pgrp %d
		"%*d "		// session %d
		"%*d "		// tty nr %d
		"%*d "		// tty pgrp %d
		"%*lu "		// flags %lu
		"%*lu "		// min flt %lu
		"%*lu "		// cmin flt %lu
		"%*lu "		// maj flt %lu
		"%*lu "		// cmaj flt %lu
		"%lu "		// utime %lu
		"%lu "		// stime %lu
		"%ld "		// cutime %ld
		"%ld "		// cstime %ld
		"%*ld "		// priority %ld
		"%*ld "		// nice %ld
		"%*ld "		// removed %ld
		"%*ld "		// itrealvalue %ld
		"%lu "		// start time %lu
		"%*lu "		// vsize %lu
		"%*ld "		// rss %ld
		"%*lu "		// rlim %lu
		"%*lu "		// start code %lu
		"%*lu "		// end code %lu
		"%*lu "		// start stack %lu
		"%*lu "		// esp %lu
		"%*lu "		// eip %lu
		"%*lu "		// signal %lu
		"%*lu "		// blocked %lu
		"%*lu "		// sigignore %lu
		"%*lu "		// sigcatch %lu
		"%*lu "		// wchan %lu
		"%*lu "		// nswap %lu
		"%*lu "		// cnswap %lu
		"%*d "		// exit signal %d
		"%*d";		// cpu number %d

	unsigned long utime, stime, cutime, cstime, start_time;
	uint64_t total_time, seconds;
	unsigned pcpu;
	time_t secs_since_boot;
	FILE* fd = _openStatFile(pid);

	if(!fd) 
	{
		return 0.0;
	}

	fscanf(fd, fmt, &utime, &stime, &cutime, &cstime, &start_time);
	fclose(fd);

	total_time = utime + stime;

	if(includeDeadChildren) 
	{
		total_time += cutime + cstime;
	}

	pcpu = 0;
	secs_since_boot = time(NULL) - cmpiutilGetSystemBootTime();
	seconds = secs_since_boot - start_time / HERTZ;
	if(seconds) 
	{
		pcpu = (total_time * 1000ULL / HERTZ) / seconds;
	}

	if(pcpu > 999U) 
	{
		pcpu = 999U;
	}

	return (double)pcpu / 10.0;
}

/*****************************************************************************/
const char*
cmpiutilGetProcCommandLine(
	pid_t pid,
	char* cmdLineOut,
	size_t cmdLineOutLen)
{
	int fd, len, i;
	char fname[24], argbfr[1024];

	if (!cmdLineOut || !cmdLineOutLen) 
	{
		errno = EINVAL;
		return NULL;
	}

	*cmdLineOut = 0;
	snprintf(fname, sizeof(fname), "/proc/%d/cmdline", pid);
	fd = open(fname, O_RDONLY);
	if (fd == -1) 
	{
		return NULL;
	}

	len = read(fd, argbfr, sizeof (argbfr));
	close(fd);
	argbfr[len] = 0;
	for(i = 0; i < len; i++) 
	{
		if (!argbfr[i])
			argbfr[i] = ' ';
	}

	return cmpiutilStrNCpy(cmdLineOut, argbfr, cmdLineOutLen);
}

/*****************************************************************************/
const char*
cmpiutilGetProcRootDir(
	pid_t pid,
	char* rootDirOut,
	size_t rootDirOutLen)
{
	char bfr[PATH_MAX+1];
	char ename[32];
	int count;

	if (!rootDirOut || !rootDirOutLen) 
	{
		errno = EINVAL;
		return NULL;
	}

	snprintf(ename, sizeof(ename), "/proc/%d/root", pid);
	count = readlink(ename, bfr, sizeof(bfr)-1);
	if(count == -1) 
	{
		return NULL;
	}

	bfr[count] = 0;
	return cmpiutilStrNCpy(rootDirOut, bfr, rootDirOutLen);
}

/*****************************************************************************/
const char*
cmpiutilGetProcCwd(
	pid_t pid,
	char* cwdOut,
	size_t cwdOutLen)
{
	char bfr[PATH_MAX+1], dirname[32];
	int count;

	if (!cwdOut || !cwdOutLen) 
	{
		errno = EINVAL;
		return NULL;
	}

	snprintf(dirname, sizeof(dirname), "/proc/%d/cwd", pid);
	count = readlink(dirname, bfr, sizeof(bfr)-1);
	if(count == -1) 
	{
		return NULL;
	}

	bfr[count] = 0;
	return cmpiutilStrNCpy(cwdOut, bfr, cwdOutLen);
}

/*****************************************************************************/
const char*
cmpiutilGetProcModulePath(
	pid_t pid,
	char* modPathOut,
	size_t modPathOutLen)
{
	char bfr[PATH_MAX+1], ename[32];
	int count;

	if (!modPathOut || !modPathOutLen) 
	{
		errno = EINVAL;
		return NULL;
	}

	snprintf(ename, sizeof(ename), "/proc/%d/exe", pid);
	count = readlink(ename, bfr, sizeof(bfr)-1);
	if(count == -1) 
	{
		/* Maybe a kernel process? */
		return cmpiutilGetProcExecName(pid, modPathOut, modPathOutLen);
	}

	bfr[count] = 0;
	return cmpiutilStrNCpy(modPathOut, bfr, modPathOutLen);
}

/*****************************************************************************/
pid_t
cmpiutilGetProcParentId(
	pid_t pid)
{
	pid_t ppid = -1;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char* fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &ppid);
		fclose(fd);
	}

	return ppid;
}

/*****************************************************************************/
int
cmpiutilGetProcGroup(
	pid_t pid)
{
	int pgrp = -1;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char* fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &pgrp);
		fclose(fd);
	}

	return pgrp;
}

/*****************************************************************************/
int
cmpiutilGetProcTTYNumber(
	pid_t pid)
{
	int ttynumber = -1;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char* fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &ttynumber);
		fclose(fd);
	}

	return ttynumber;
}

/*****************************************************************************/
int
cmpiutilGetProcSessionID(
	pid_t pid)
{
	int sessionid = -1;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char* fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &sessionid);
		fclose(fd);
	}

	return sessionid;
}

/*****************************************************************************/
int
cmpiutilGetProcOwningPGRPOfTTY(
	pid_t pid)
{
	int ttypgrp = -1;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char* fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &ttypgrp);
		fclose(fd);
	}

	return ttypgrp;
}

/*****************************************************************************/
unsigned long
cmpiutilGetProcFlags(
	pid_t pid)
{
	unsigned long flags = (unsigned long) -1;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char *fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &flags);
		fclose(fd);
	}
	return flags;
}

/*****************************************************************************/
long
cmpiutilGetProcPriority(
	pid_t pid)
{
	long priority= -1L;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char *fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &priority);
		fclose(fd);
		priority -= 15;
	}

	return priority;
}

/*****************************************************************************/
uint64_t
cmpiutilGetProcKernelModeTime(
	pid_t pid)
{
	uint64_t kmtime = 0;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char *fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%llu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &kmtime);
		fclose(fd);
		kmtime *= 10;
	}

	return kmtime;
}

/*****************************************************************************/
uint64_t
cmpiutilGetUserModeTime(
	pid_t pid)
{
	uint64_t umtime = 0;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char *fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%llu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &umtime);
		fclose(fd);
		umtime *= 10;
	}

	return umtime;
}

/*****************************************************************************/
long
cmpiutilGetProcNiceValue(
	pid_t pid)
{
	long nice = -1L;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char *fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &nice);
		fclose(fd);
	}
	return nice;
}

/*****************************************************************************/
time_t
cmpiutilGetProcStartTime(
	pid_t pid)
{
	time_t rc = 0;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		unsigned long starttime;
		const char *fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%lu "		// start time %lu
			"%*lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &starttime);
		fclose(fd);
		rc = (starttime / 100) + cmpiutilGetSystemBootTime();
	}

	return rc;
}

/*****************************************************************************/
unsigned long
cmpiutilGetProcVirtualMemSize(
	pid_t pid)
{
	unsigned long vsize = 0;
	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		const char *fmt =
			"%*d "		// pid %d
			"%*s "		// comm %s
			"%*c "		// state %c
			"%*d "		// ppid %d
			"%*d "		// pgrp %d
			"%*d "		// session %d
			"%*d "		// tty nr %d
			"%*d "		// tty pgrp %d
			"%*lu "		// flags %lu
			"%*lu "		// min flt %lu
			"%*lu "		// cmin flt %lu
			"%*lu "		// maj flt %lu
			"%*lu "		// cmaj flt %lu
			"%*lu "		// utime %lu
			"%*lu "		// stime %lu
			"%*ld "		// cutime %ld
			"%*ld "		// cstime %ld
			"%*ld "		// priority %ld
			"%*ld "		// nice %ld
			"%*ld "		// removed %ld
			"%*ld "		// itrealvalue %ld
			"%*lu "		// start time %lu
			"%lu "		// vsize %lu
			"%*ld "		// rss %ld
			"%*lu "		// rlim %lu
			"%*lu "		// start code %lu
			"%*lu "		// end code %lu
			"%*lu "		// start stack %lu
			"%*lu "		// esp %lu
			"%*lu "		// eip %lu
			"%*lu "		// signal %lu
			"%*lu "		// blocked %lu
			"%*lu "		// sigignore %lu
			"%*lu "		// sigcatch %lu
			"%*lu "		// wchan %lu
			"%*lu "		// nswap %lu
			"%*lu "		// cnswap %lu
			"%*d "		// exit signal %d
			"%*d";		// cpu number %d

		fscanf(fd, fmt, &vsize);
		fclose(fd);
	}

	return vsize;
}

/*****************************************************************************/
unsigned long
cmpiutilGetProcResidentSetSize(
	pid_t pid)
{
	unsigned long residentSetSize = 0;
	FILE *fd = _openMemFile(pid);
	if(fd) 
	{
		const char* fmt = "%*lu %lu %*lu %*lu %*lu %*lu %*lu";
		fscanf(fd, fmt, &residentSetSize);
		fclose(fd);
		residentSetSize *= 4096;
	}

	return residentSetSize;
}

/*****************************************************************************/
unsigned long
cmpiutilGetProcSharedMemSize(
	pid_t pid)
{
	unsigned long sharedMemSize = 0;
	FILE* fd = _openMemFile(pid);
	if(fd) 
	{
		const char* fmt = "%*lu %*lu %lu %*lu %*lu %*lu %*lu";
		fscanf(fd, fmt, &sharedMemSize);
		fclose(fd);
		sharedMemSize *= 4096;
	}

	return sharedMemSize;
}

/*****************************************************************************/
int
cmpiutilGetProcStats(
	pid_t pid,
	process_stats_t* processStatsOut)
{
	const char *fmt;

	if (!processStatsOut) 
	{
		errno = EINVAL;
		return -1;
	}

	memset(processStatsOut, 0, sizeof(*processStatsOut));
	if (cmpiutilGetProcExecName(pid, processStatsOut->execName,
		sizeof(processStatsOut->execName)) == NULL) 
	{
		return -1;
	}

	fmt =
		"%*d "      // pid %d
		"%*s "       // --> comm %s
		"%c "		// --> state %c
		"%d "		// --> ppid %d
		"%d "		// --> pgrp %d
		"%d "		// --> session %d
		"%d "		// --> tty nr %d
		"%*d "		// tty pgrp %d
		"%*lu "		// flags %lu
		"%*lu "		// min flt %lu
		"%*lu "		// cmin flt %lu
		"%*lu "		// maj flt %lu
		"%*lu "		// cmaj flt %lu
		"%lu "		// --> utime %lu
		"%lu "		// --> stime %lu
		"%ld "		// --> cutime %ld
		"%ld "		// --> cstime %ld
		"%ld "		// --> priority %ld
		"%ld "		// --> nice %ld
		"%*ld "		// removed %ld
		"%*ld "		// itrealvalue %ld
		"%lu "		// --> start time %lu
		"%lu "		// --> vsize %lu
		"%*ld "		// rss %ld
		"%*lu "		// rlim %lu
		"%*lu "		// start code %lu
		"%*lu "		// end code %lu
		"%*lu "		// start stack %lu
		"%*lu "		// esp %lu
		"%*lu "		// eip %lu
		"%*lu "		// signal %lu
		"%*lu "		// blocked %lu
		"%*lu "		// sigignore %lu
		"%*lu "		// sigcatch %lu
		"%*lu "		// wchan %lu
		"%*lu "		// nswap %lu
		"%*lu "		// cnswap %lu
		"%*d "		// exit signal %d
		"%*d";		// cpu number %d

	FILE* fd = _openStatFile(pid);
	if(fd) 
	{
		uint64_t total_time, seconds;
		uint32_t pcpu;
		time_t secs_since_boot;
		char    state;
		uint32_t utime, stime, cstime, cutime;
		unsigned long starttime;

		fscanf(fd, fmt,
				&state,
				&processStatsOut->parentProcessID,
				&processStatsOut->processGroup,
				&processStatsOut->sessionID,
				&processStatsOut->ttyNumber,
				&utime,
				&stime,
				&cutime,
				&cstime,
				&processStatsOut->priority,
				&processStatsOut->niceValue,
				&starttime,
				&processStatsOut->virtualMemSize
				);
		fclose(fd);

		processStatsOut->state = _stateThruTasks(pid);
		processStatsOut->startTime = (starttime/100) + cmpiutilGetSystemBootTime();
		processStatsOut->kernelModeTime = stime;
		processStatsOut->userModeTime   = utime;

		/* parse percent CPU utilization... */
		total_time = utime + stime;

#if 0
		if(includeDeadChildren) 
		{
			total_time += cutime + cstime;
		}
#endif

		pcpu = 0;
		secs_since_boot = time(NULL) - _system_boot_time;
		seconds = secs_since_boot - starttime / HERTZ;
		if(seconds) 
		{
			if((pcpu = (total_time * 1000ULL / HERTZ) / seconds) > 999U) 
			{
				pcpu = 999U;
			}
		}

		processStatsOut->percentCPU = (double)pcpu / 10.0;
	} /* if - stat file opened successfully */
	else 
	{
		return -1;
	}

	fd = _openMemFile(pid);
	if(fd) 
	{
		unsigned long residentSetSize, sharedMemSize;
		const char* fmt = "%*lu %lu %lu %*lu %*lu %*lu %*lu";
		fscanf(fd, fmt, &residentSetSize, &sharedMemSize);
		fclose(fd);
		processStatsOut->residentSetSize = residentSetSize * 4096;
		processStatsOut->sharedMemSize = sharedMemSize * 4096;
	}
	else 
	{
		return -1;
	}

	return 0;
}

/*****************************************************************************/
static int
_getTaskState(
	pid_t pid,
	int tid)
{
	int cc = -1;
	char stateChar;
	char fname[64];
	FILE *fd;

	const char* fmt =
		"%*d "		/* pid %d */
		"%*s "		/* comm %s */
		"%c "		/* state %c */
		"%*d "		/* ppid %d */
		"%*d "		/* pgrp %d */
		"%*d "		/* session %d */
		"%*d "		/* tty nr %d */
		"%*d "		/* tty pgrp %d */
		"%*lu "		/* flags %lu */
		"%*lu "		/* min flt %lu */
		"%*lu "		/* cmin flt %lu */
		"%*lu "		/* maj flt %lu */
		"%*lu "		/* cmaj flt %lu */
		"%*lu "		/* utime %lu */
		"%*lu "		/* stime %lu */
		"%*ld "		/* cutime %ld */
		"%*ld "		/* cstime %ld */
		"%*ld "		/* priority %ld */
		"%*ld "		/* nice %ld */
		"%*ld "		/* removed %ld */
		"%*ld "		/* itrealvalue %ld */
		"%*lu "		/* start time %lu */
		"%*lu "		/* vsize %lu */
		"%*ld "		/* rss %ld */
		"%*lu "		/* rlim %lu */
		"%*lu "		/* start code %lu */
		"%*lu "		/* end code %lu */
		"%*lu "		/* start stack %lu */
		"%*lu "		/* esp %lu */
		"%*lu "		/* eip %lu */
		"%*lu "		/* signal %lu */
		"%*lu "		/* blocked %lu */
		"%*lu "		/* sigignore %lu */
		"%*lu "		/* sigcatch %lu */
		"%*lu "		/* wchan %lu */
		"%*lu "		/* nswap %lu */
		"%*lu "		/* cnswap %lu */
		"%*d "		/* exit signal %d */
		"%*d";		/* cpu number %d */

	snprintf(fname, sizeof(fname), "/proc/%d/task/%d/stat", pid, tid);
	fd = fopen(fname, "r");
	if (!fd) 
	{
		return cc;
	}

	setvbuf(fd, NULL, _IONBF, 0);
	fscanf(fd, fmt, &stateChar);
	fclose(fd);
	switch(stateChar) 
	{
		case 'T': cc = 1; break;	/* Traced or stopped */
		case 'R': cc = 2; break;	/* Runnable (on run queue) */
		case 'D': cc = 3; break;	/* Uninterruptible sleep (usually i/o)*/
		case 'Z': cc = 4; break;	/* Zombie */
		case 'S': cc = 5; break;	/* Sleeping */
		case 'W': cc = 6; break;	/* Paging */
		default: cc = 7; break;		/* Unknown */
	}

	return cc;
}

/*****************************************************************************/
static enum ProcState
_stateThruTasks(
	pid_t pid)
{
	enum ProcState psra[] =
	{
		EPROCSTATE_Unknown,
		EPROCSTATE_Stopped,
		EPROCSTATE_Running,
		EPROCSTATE_Blocked,
		EPROCSTATE_Terminated,
		EPROCSTATE_SuspendedReady,
		EPROCSTATE_Other,
		EPROCSTATE_Unknown
	};

	DIR* pdir;
	struct dirent* pdirent;
	char dirname[32];
	int statepriority;
	int savedpriority = 7;

	snprintf(dirname, sizeof(dirname), "/proc/%d/task", pid);
	pdir = opendir(dirname);
	if(!pdir) 
	{
		return (int)EPROCSTATE_DoesntExist;
	}

	while((pdirent = readdir(pdir)) != NULL) 
	{
		char *endptr;
		long tid = strtol(pdirent->d_name, &endptr, 10);
		if(endptr == pdirent->d_name || *endptr) 
		{
			continue;
		}

		statepriority = _getTaskState(pid, (int)tid);
		if(statepriority > -1 && statepriority < savedpriority) 
		{
			savedpriority = statepriority;
		}
	}

	closedir(pdir);
	return (int)psra[savedpriority];
}

/*****************************************************************************/
static FILE*
_openStatFile(
	pid_t pid)
{
	char fname[32];
	FILE *fd;

	snprintf(fname, sizeof(fname), "/proc/%d/stat", pid);
	fd = fopen(fname, "r");
	if(fd) 
	{
		setvbuf(fd, NULL, _IONBF, 0);
	}
	return fd;
}

/*****************************************************************************/
static FILE*
_openMemFile(
	pid_t pid)
{
	char fname[32];
	FILE *fd;

	snprintf(fname, sizeof(fname), "/proc/%d/statm", pid);
	fd = fopen(fname, "r");
	if(fd) 
	{
		setvbuf(fd, NULL, _IONBF, 0);
	}
	return fd;
}

/*****************************************************************************/
static int
_appendPid2Array(
	pid_t** intBfr,
	size_t* intBfrSize,
	size_t* offset,
	pid_t value)
{
	if (!(*intBfr)) 
	{
		*intBfr = (pid_t*)malloc(256 * sizeof(pid_t));
		if (!(*intBfr)) 
		{
			*offset = 0;
			return -1;
		}
		*intBfrSize = 256;
		*offset = 1;
		*intBfr[0] = value;
		return 0;
	}

	if (*offset >= *intBfrSize) 
	{
		pid_t* lbfr = (pid_t*)realloc(*intBfr, (*intBfrSize)+128);
		if (!lbfr) 
		{
			free (*intBfr);
			*intBfr = 0;
			*offset = 0;
			return -1;
		}
		*intBfrSize += 128;
		*intBfr = lbfr;
	}
	(*intBfr)[*offset] = value;
	(*offset)++;
	return 0;
}

