/*******************************************************************************
* Copyright (C) 2005, 2006 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
* Author: Norm Paxton (npaxton@novell.com)
******************************************************************************/

#include <cmpiutil/ntpServer.h>
#include <cmpiutil/base.h>
#include <cmpiutil/ioUtils.h>
#include <cmpiutil/string.h>
#include <cmpiutil/modifyFile.h>

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include <pthread.h>

#define LOCAL_DEBUG

#ifdef LOCAL_DEBUG
#define DEBUGOUT(fmt, args...) printf(fmt,## args)
#else
#define DEBUGOUT(fmt, args...)
#endif


static pthread_mutex_t fileGuard = PTHREAD_MUTEX_INITIALIZER;
NTPServer ntpServer = {{0},0,0,0,0};

#ifdef __cplusplus
extern "C" {
#endif



int
serverSupportsNTP()
{
	unsigned long installDate;
	installDate = cmpiutilGetRPMInstallDate("xntp");
	if (installDate)
	{
		return 1;
	}
	return 0;
}


char *
lineFromServer(
	NTPServer *pServer)
{
	char *line = (char *)malloc(256);
	char temp[16] = {0};

	if (line)
	{
		memset(line, 0, 256);

		cmpiutilStrAppend(line, "server ", 256);
		cmpiutilStrAppend(line, pServer->serverName, 256);

		if(pServer->prefer)
		{
			cmpiutilStrAppend(line, " prefer", 256);
		}
		if(pServer->key)
		{
			cmpiutilStrAppend(line, " key ", 256);
			sprintf(temp, "%u", pServer->key);
			cmpiutilStrAppend(line, temp, 256);
		}
		if(pServer->ntpVersion)
		{
			cmpiutilStrAppend(line, " version ", 256);
			sprintf(temp, "%u", pServer->ntpVersion);
			cmpiutilStrAppend(line, temp, 256);
		}
		if(pServer->minPoll)
		{
			cmpiutilStrAppend(line, " minpoll ", 256);
			sprintf(temp, "%u", pServer->minPoll);
			cmpiutilStrAppend(line, temp, 256);
		}
		if(pServer->maxPoll)
		{
			cmpiutilStrAppend(line, " maxpoll ", 256);
			sprintf(temp, "%u", pServer->maxPoll);
			cmpiutilStrAppend(line, temp, 256);
		}
	}

	DEBUGOUT("  lineFromServer returning: %s\n", line);
	return line;
}


static void
serverFromLine(
	const char *line,
	NTPServer *pServer)
{
	if (!pServer)
		return;

	DEBUGOUT("  into serverFromLine with line: %s\n", line);

	unsigned int i = 0;
	size_t iTokenCount = 0;
	char** tokens = cmpiutilStrTokenize(line, " ,", &iTokenCount);

	for(i = 0; i < iTokenCount; i++)
	{
		DEBUGOUT("  token: %d: %s\n", i, tokens[i]);
	}

	if(iTokenCount > 1)
	{
		// strip serverName... including # from end
		strncpy(pServer->serverName, tokens[1], NTP_SERVER_NAME_LEN);
		while(pServer->serverName[strlen(pServer->serverName)-1]=='#')
		{
			pServer->serverName[strlen(pServer->serverName)-1]='\0';
		}
		cmpiutilStrRTrim(pServer->serverName);
		if(iTokenCount > 2)
		{
			DEBUGOUT("  iTokenCount: %d\n", (int)iTokenCount);
			for(i = 2; i < iTokenCount; i++)
			{
				if(strcasecmp(tokens[i],"prefer") == 0)
				{
					pServer->prefer = 1;
					DEBUGOUT("  prefer: %d\n", pServer->prefer);
					continue;
				}
				if(i < iTokenCount-1)
				{
					if(strcasecmp(tokens[i],"key") == 0)
					{
						i++;
						pServer->key = (unsigned long)atol(tokens[i]);
						DEBUGOUT("  key: %d\n", pServer->key);
					}
					else if(strcasecmp(tokens[i],"version") == 0)
					{
						i++;
						pServer->ntpVersion = (unsigned long)atol(tokens[i]);
						DEBUGOUT("  ntpVersion: %d\n", pServer->ntpVersion);
					}
					else if(strcasecmp(tokens[i],"mode") == 0)
					{
						i++;
						// Ignore
					}
					else if(strcasecmp(tokens[i],"minpoll") == 0)
					{
						i++;
						pServer->minPoll = (unsigned long)atol(tokens[i]);
						DEBUGOUT("  minpoll: %d\n", pServer->minPoll);
					}
					else if(strcasecmp(tokens[i],"maxpoll") == 0)
					{
						i++;
						pServer->maxPoll = (unsigned long)atol(tokens[i]);
						DEBUGOUT("  maxpoll: %d\n", pServer->maxPoll);
					}
				}
			}
		}
	}
}


/****************************************************************************/
static int
errCond(
	char* errMsgBfr,
	unsigned int errMsgBfrLen,
	int cc,
	const char* fmt,
	...)
{
	va_list ap;

	if (!errMsgBfr || !errMsgBfrLen) {
		return cc;
	}

	va_start(ap, fmt);
	vsnprintf(errMsgBfr, errMsgBfrLen, fmt, ap);
	va_end(ap);
	return cc;
}



/*****************************************************************************/
/* NOTE: This function assumes mutex is already locked */
static int myNTPServerModifyLine(	const char* line,
				FILE* const ostrm,
				char* errMsgBfr,
				size_t errMsgBfrLen)
{
	DEBUGOUT("myNTPServerModifyLine() called\n");

	int bMod = 0;
	NTPServer lookupServer = {{0},0,0,0,0};

	if (cmpiutilStrStartsWith(line, "server"))
	{
		serverFromLine(line, &lookupServer);
		if(strcasecmp(lookupServer.serverName, ntpServer.serverName) == 0)
		{
			DEBUGOUT("  myNTPServerModifyLine() about to write out ntpServer: \n   name: %s\n   key: %d\n   version: %d\n   minpoll: %d\n   maxpoll: %d\n", ntpServer.serverName, ntpServer.key, ntpServer.ntpVersion, ntpServer.minPoll, ntpServer.maxPoll);
			char *newLine = lineFromServer(&ntpServer);
			DEBUGOUT("  myNTPServerModifyLine() modifying line for server: %s\n   oldline: %s\n   newline: %s\n", ntpServer.serverName, line, newLine);
			fputs(newLine, ostrm);
			fputs("\n", ostrm);
			bMod = 1;
			free(newLine);
		}
	}

	if (!bMod)
	{
		// didn't modify this line, just write it back out...
		// bModify is still 0, let end return handle it
		fputs(line, ostrm);
		fputs("\n", ostrm);
	}

	DEBUGOUT("myNTPServerModifyLine() returning\n");

	return bMod;
}


/*****************************************************************************/
/* NOTE: This function assumes mutex is already locked */
static int myNTPServerDeleteLine(	const char* line,
				FILE* const ostrm,
				char* errMsgBfr,
				size_t errMsgBfrLen)
{
	DEBUGOUT("myNTPServerDeleteLine() called\n");

	int bDel = 0;
	NTPServer server = {{0},0,0,0,0};

	if (cmpiutilStrStartsWith(line, "server"))
	{
		serverFromLine(line, &server);
		if(strcasecmp(server.serverName, ntpServer.serverName) == 0)
		{
			DEBUGOUT("myNTPServerDeleteLine() removing line for server: %s\n   %s\n", ntpServer.serverName, line);
			bDel = 1;
		}
	}

	if (!bDel)
	{
		// didn't modify this line, just write it back out...
		// bModify is still 0, let end return handle it
		fputs(line, ostrm);
		fputs("\n", ostrm);
	}

	DEBUGOUT("myNTPServerDeleteLine() returning\n");

	return bDel;
}


/*****************************************************************************/
// int iDel  1 = Delete,  0 = Modify
static int modifyNTPServer(const char *serverName,
							NTPServer *pServer,
							int iDel)
{
	DEBUGOUT("modifyNTPServer() called\n");

	int bModified = 0;
	int cc = pthread_mutex_lock(&fileGuard);

	if (pServer && *(pServer->serverName))
	{
		copyNTPServer(pServer, &ntpServer);
		DEBUGOUT("  modifyNTPServer() setting ntpServer: %s\n   key: %d\n   version: %d\n   minpoll: %d\n   maxpoll: %d\n", ntpServer.serverName, ntpServer.key, ntpServer.ntpVersion, ntpServer.minPoll, ntpServer.maxPoll);
	}
	else
	{
		strncpy(ntpServer.serverName, serverName, NTP_SERVER_NAME_LEN);
	}


	cmpiutilFileModifier_CBS fms = {0};
	if (iDel)
	{
		fms.writeLine = myNTPServerDeleteLine;
	}
	else
	{
		fms.writeLine = myNTPServerModifyLine;
	}

	char errmsg[256] = {0};

	int r1 = cmpiutilModifyFile("/etc/ntp.conf",
						&fms,
						errmsg,
						256);

	if (r1)
	{
		DEBUGOUT("Modifying file %s returned error:\n  Attempting rollback\n",
			     errmsg);
		r1 = cmpiutilModifyRollBack("/etc/ntp.conf",
								errmsg,
								256 );
		if (r1)
		{
			DEBUGOUT("Rollback failed.  use .orig file for /etc/ntp.conf\n err1: %s\n",
					 errmsg);
		}
	}

	cc = pthread_mutex_unlock(&fileGuard);

	DEBUGOUT("modifyNTPServer() returning\n");

	return bModified;
}


/*****************************************************************************/
NTPServer *getNTPServers(int *pCount)
{
	FILE *file;
	size_t bfrLen = 0;
	char *p = NULL, *bufPtr = NULL;
	NTPServer server = {{0},0,0,0,0};
	char* errMsgBfr = NULL;
	unsigned int errMsgBfrLen = 0;
	int i=0;
	int iCount = 0;

	DEBUGOUT("Entering getNTPServers\n");

	if (!pCount)
	{
		return 0;
	}

	file = fopen("/etc/ntp.conf", "r");
	if (!file)
	{
		errCond(errMsgBfr, errMsgBfrLen, -1,
			"Failed to open file /etc/ntp.conf");
		return NULL;
	}

	for(p = cmpiutilReadLine(file, &bufPtr, &bfrLen); p;
		p = cmpiutilReadLine(file, &bufPtr, &bfrLen))
	{
		DEBUGOUT("  got line: %s\n", bufPtr);

		if (cmpiutilStrStartsWith(p, "server"))
		{
			iCount++;
		}
	}

	*pCount = iCount;
	DEBUGOUT("  malloc'ing for %d servers in ra\n", iCount);
	NTPServer *ra = (NTPServer*)malloc(iCount * sizeof(NTPServer));

	// reset file pointer
	rewind(file);

	i = 0;
	for(p = cmpiutilReadLine(file, &bufPtr, &bfrLen); p;
		p = cmpiutilReadLine(file, &bufPtr, &bfrLen))
	{
		DEBUGOUT("  got line: %s\n", bufPtr);

		if (cmpiutilStrStartsWith(p, "server"))
		{
			serverFromLine(p, &server);
			DEBUGOUT("  Resulted in NTPServer in list: serverName: %s\n", server.serverName);
			copyNTPServer(&server, &ra[i]);

			DEBUGOUT("  Adding NTPServer: %s\n", ra[i].serverName);
			i++;
		}
	}

	fclose(file);

	if (bufPtr)
		free(bufPtr);

	return ra;
}


/*****************************************************************************/
NTPServer *getNTPServer(const char *serverName)
{
	FILE *file;
	size_t bfrLen = 0;
	char *p, *bufPtr = NULL;
	NTPServer server = {{0},0,0,0,0};
	char* errMsgBfr = NULL;
	unsigned int errMsgBfrLen = 0;

	if (!serverName)
	{
		return NULL;
	}

	file = fopen("/etc/ntp.conf", "r");
	if (!file)
	{
		errCond(errMsgBfr, errMsgBfrLen, -1,
			"Failed to open file /etc/ntp.conf");
		return NULL;
	}

	for(p = cmpiutilReadLine(file, &bufPtr, &bfrLen); p;
		p = cmpiutilReadLine(file, &bufPtr, &bfrLen))
	{
		if (cmpiutilStrStartsWith(p, "server"))
		{
			serverFromLine(p, &server);
			if(strcasecmp(server.serverName, serverName) == 0)
			{
				return createAndCopyNTPServer(&server);
			}
		}
	}

	return NULL;
}


/*****************************************************************************/
int deleteNTPServer(const char *serverName)
{
	return modifyNTPServer(serverName, NULL, 1); // delete line
}


/*****************************************************************************/
int setNTPServer(const char *serverName, NTPServer *server)
{
	return modifyNTPServer(serverName, server, 0); // modify line
}


/*****************************************************************************/
void copyNTPServer(const NTPServer *in, NTPServer *out)
{
	if (in && out)
	{
		strncpy(out->serverName, in->serverName, NTP_SERVER_NAME_LEN);
		out->prefer = in->prefer;
		out->key = in->key;
		out->ntpVersion = in->ntpVersion;
		out->minPoll = in->minPoll;
		out->maxPoll = in->maxPoll;
	}
}


/*****************************************************************************/
NTPServer *createAndCopyNTPServer(const NTPServer* in)
{
	if (!in)
		return NULL;
	NTPServer *out = (NTPServer *)malloc(sizeof(NTPServer));
	if (out)
	{
		copyNTPServer(in, out);
	}
	return out;
}

